@@ -1,13 +1,13 @@
 /* spell.c -- clone of Unix `spell'.
 
    This file is part of GNU Spell.
-   Copyright (C) 1996 Free Software Foundation, Inc.
-   Written by Thomas Morgan <tmorgan@pobox.com>
+   Copyright (C) 1996,2010 Free Software Foundation, Inc.
+   Written by Thomas Morgan <tmorgan@pobox.com> and others
 
-   This program is free software; you can redistribute it and/or modify
+   This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; either version 2, or (at your option)
-   any later version.
+   the Free Software Foundation, either version 3 of the License, or
+   (at your option) any later version.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -15,8 +15,8 @@
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
-   along with this program; if not, write to the Free Software
-   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */
+   along with this program.  If not, see <http://www.gnu.org/licenses/>.
+*/
 
 /* Local headers.  */
 
@@ -24,9 +24,6 @@
 #include "config.h"
 #endif
 
-#include "getopt.h"
-#include "str.h"
-
 /* System headers.  */
 
 #include <sys/types.h>
@@ -42,6 +39,7 @@
 #include <sys/stat.h>
 #include <sys/time.h>
 #include <unistd.h>
+#include <getopt.h>
 
 #ifdef HAVE_STRING_H
 #include <string.h>
@@ -72,8 +70,10 @@ struct pipe
   {
     /* File descriptors used by the parent process.  */
     int pin;			/* Input channel.  */
+    FILE* fpin;
     int pout;			/* Output channel.  */
     int perr;			/* Error channel (for reading).  */
+    FILE* fperr;
 
     /* File descriptors used by the child process.  */
     int cin;			/* Input channel.  */
@@ -89,7 +89,6 @@ typedef struct pipe pipe_t;
 static char *strerror (int);
 #endif
 
-char *find_ispell ();
 static char *xstrdup (const char *);
 static void *xmalloc (size_t);
 static void *xrealloc (void *, size_t);
@@ -103,9 +102,6 @@ void read_ispell (pipe_t *, char *, int);
 void read_ispell_errors (pipe_t *);
 void run_ispell_in_child (pipe_t *);
 
-/* Version of this program.  */
-const char version[] = "version " VERSION;
-
 /* Switch information for `getopt'.  */
 const struct option long_options[] =
 {
@@ -127,7 +123,7 @@ const struct option long_options[] =
 /* The name of the executable this process comes from.  */
 char *program_name = NULL;
 
-/* Ispell's location.  */
+/* Location of Ispell/Aspell.  */
 char *ispell_prog = NULL;
 
 /* Dictionary to use.  Just use the default if NULL.  */
@@ -159,7 +155,7 @@ int interactive = 0;
 int
 main (int argc, char **argv)
 {
-  char opt = 0;			/* Current option.  */
+  int opt = 0;			/* Current option.  */
   int opt_error = 0;		/* Whether an option error occurred.  */
   int show_help = 0;		/* Display help (--help, -h).  */
   int show_version = 0;		/* Display the version (--version, -V).  */
@@ -232,16 +228,19 @@ main (int argc, char **argv)
 
   if (show_version)
     {
-      error (0, 0, version);
-
-      if (!show_help)
-	exit (EXIT_SUCCESS);
+      puts ( "GNU Spell " VERSION "\n"
+	"Copyright (C) 1996,2010 Free Software Foundation, Inc.\n"
+	"License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>\n"
+	"This is free software: you are free to change and redistribute it.\n"
+	"There is NO WARRANTY, to the extent permitted by law.\n"
+	);
+      exit (EXIT_SUCCESS);
     }
 
   if (show_help)
     {
       printf ("Usage: %s [OPTION]... [FILE]...\n", program_name);
-      fputs ("This is GNU Spell, a Unix spell emulator.\n\n"
+      puts ("This is GNU Spell, a Unix spell emulator.\n\n"
 	     "  -I, --ispell-version\t\tPrint Ispell's version.\n"
 	     "  -V, --version\t\t\tPrint the version number.\n"
 	     "  -b, --british\t\t\tUse the British dictionary.\n"
@@ -254,12 +253,16 @@ main (int argc, char **argv)
 	     "  -s, --stop-list=FILE\t\tIgnored; for compatibility.\n"
 	     "  -v, --verbose\t\t\tPrint words not literally found.\n"
 	     "  -x, --print-stems\t\tIgnored; for compatibility.\n\n"
-	     "Please use Info to read more (type `info spell').\n", stderr);
+	     "Please use Info to read more (type `info spell').\n"
+	     "Report bugs to: cate@gnu.org\n"
+	     "spell home page: <http://www.gnu.org/software/spell/>\n"
+	     "General help using GNU software: <http://www.gnu.org/gethelp/>\n"
+	   );
       exit (EXIT_SUCCESS);
     }
 
   if (!ispell_prog)
-    ispell_prog = find_ispell ();
+    ispell_prog = "ispell";
 
   new_pipe (&ispell_pipe);
 
@@ -275,91 +278,75 @@ main (int argc, char **argv)
   exit (EXIT_SUCCESS);
 }
 
-/* Return the location of Ispell through the string *ISPELL (created
-   by `str_make'), or find it in the `PATH' environmental variable,
-   or exit with an error if it is not found.  */
-
-char *
-find_ispell ()
-{
-  char *ispell = NULL;
-  char *path = NULL;
-  int path_len = 0;
-  int pos = 0;
-  str_t *file = str_make (0);
-  struct stat stat_buf;
-
-  path = xstrdup (getenv ("PATH"));
-  path_len = strlen (path);
-
-  while (1)
-    {
-      file = str_make (file);
-
-      if (!pos && path[pos] == ':')
-	/* A `:' right at the beginning means the current
-	   directory is the first directory to search.  */
-	str_add_char (file, '.');
-      else
-	for (; pos < path_len && path[pos] != ':'; pos++)
-	  str_add_char (file, path[pos]);
-
-      if (file->str[file->len - 1] != '/')
-	str_add_char (file, '/');
-      str_add_str (file, nstr_to_str ("ispell"));
-
-      if (stat (str_to_nstr (file), &stat_buf) != -1)
-	return xstrdup (str_to_nstr (file));
-
-      if (pos >= path_len)
-	error (EXIT_FAILURE, 0, "unable to locate Ispell");
-      pos++;
-    }
-
-  /* We should never reach this.  */
-  abort ();
-}
 
 /* Read the file *FILE, opened in the file stream *STREAM.  Send
    output, line by line, through *THE_PIPE (created by `new_pipe').  */
+#undef INITIAL_BUFF_SIZE
+#define INITIAL_BUFF_SIZE   2048
 
 void
 read_file (pipe_t * the_pipe, FILE * stream, char *file)
 {
-  str_t *str = str_make (0);
-  enum add_line_return add_line_ret = 0;
-  int line = 0;
+  size_t buff_size = INITIAL_BUFF_SIZE;
+  char *buff = xmalloc(sizeof(char[INITIAL_BUFF_SIZE]));
+  ssize_t len, rlen, zlen;
+  char *wbuff;
+  unsigned long line = 0;
 
   while (1)
     {
-      str = str_make (str);
-
-      str_add_char (str, '^');
-      add_line_ret = str_add_line (str, stream);
+      len = getline(&buff, &buff_size, stream);
       line++;
-
-      if (add_line_ret == ADD_LINE_ERR)
-	error (EXIT_FAILURE, errno, "%s: error reading line", file);
-      if (add_line_ret == ADD_LINE_EOF && !str->len)
-	return;
-
-      /* In case there was no newline at the end of the file.  */
-      if (str->str[str->len - 1] != '\n')
-	str_add_char (str, '\n');
-
-      if (write (the_pipe->pout, str_to_nstr (str), str->len) != str->len)
-	error (EXIT_FAILURE, errno, "error writing to Ispell");
-
-      read_ispell_errors (the_pipe);
-      read_ispell (the_pipe, file, line);
-      read_ispell_errors (the_pipe);
-
-      if (add_line_ret == ADD_LINE_EOF)
-	return;
+      if (len < 0) {
+        if (feof(stream)) {
+	    free(buff);
+            return;
+	} else {
+	    error (EXIT_FAILURE, errno, "%s: error reading line", file);
+	}
+     }
+    /* In case there was no newline at the end of the file.  */
+    if (len>0  && buff[len-1] != '\n') {
+      if (len >= buff_size-1)
+	buff = xrealloc(buff, buff_size += 16);
+      buff[len] = '\n';
+      len++;
     }
+    /* The '\0' chars in the string are transformed to a space ' ' */
+    wbuff = buff;
+    rlen = 0;
+    zlen = 0;
+    while (1) {
+      rlen = strlen(wbuff);
+      zlen += rlen;
+      if (zlen >= len)
+	break;
+      wbuff[rlen] = ' ';
+      wbuff += rlen;
+    }
+
+    rlen = write (the_pipe->pout, "^", 1);
+    if (rlen <= 0)
+        error (EXIT_FAILURE, errno, "error writing to Ispell");
+
+    wbuff = buff;
+    while (len > 0) {
+      rlen = write (the_pipe->pout, wbuff, len);
+      if (rlen <= 0)
+	error (EXIT_FAILURE, errno, "error writing to Ispell");
+      len -= rlen;
+      wbuff += rlen;
+    }
+    fsync(the_pipe->pout);
+
+    read_ispell_errors (the_pipe);
+    read_ispell (the_pipe, file, line);
+    read_ispell_errors (the_pipe);
+  }
 
   if (fclose (stream) == EOF)
     error (0, errno, "%s: close error", file);
+  free(buff);
 }
 
 /* Read all of Ispell's corrections for a line of text (already
@@ -367,32 +354,41 @@ read_file (pipe_t * the_pipe, FILE * stream, char *file)
    Must be called from the parent process communicating with Ispell.
    Print out the misspelled words, processing until seeing a blank
    line.  */
+#undef INITIAL_BUFF_SIZE
+#define INITIAL_BUFF_SIZE   256
 
 void
 read_ispell (pipe_t * ispell_pipe, char *file, int line)
 {
-  str_t *str = str_make (0);
+  size_t buff_size = INITIAL_BUFF_SIZE;
+  char *buff = xmalloc(sizeof(char[INITIAL_BUFF_SIZE]));
+  ssize_t len;
+  char b0;
 
   while (1)
     {
-      str = str_make (str);
-
-      if (str_add_line_from_desc (str, ispell_pipe->pin) == ADD_LINE_EOF)
-	exit (EXIT_SUCCESS);
-
-      /* Ispell gives us a blank line when it's finished processing
-         the line we just gave it.  */
-      if (str->len == 1 && str->str[0] == '\n')
+      len = getline(&buff, &buff_size, ispell_pipe->fpin);
+      if (len == -1) {
+	if (feof(ispell_pipe->fpin))
+	    exit (EXIT_SUCCESS);
+	else
+	    error(EXIT_FAILURE, errno, "error reading data from ispell/aspell");
+     }
+     b0 = buff[0];
+     /* Ispell gives us a blank line when it's finished processing
+        the line we just gave it.  */
+     if (len == 1 && b0 == '\n') {
+      free(buff);
 	return;
-
+      }
       /* There was no problem with this word.  */
-      if (str->str[0] == '*' || str->str[0] == '+'
-	  || str->str[0] == '-')
+      if (b0 == '*' || b0 == '+'
+	  || b0 == '-')
 	continue;
 
       /* The word appears to have been misspelled.  */
-      if (str->str[0] == '&' || str->str[0] == '#'
-	  || (str->str[0] == '?' && verbose))
+      if (b0 == '&' || b0 == '#'
+	  || (b0 == '?' && verbose))
 	{
 	  int pos;
 
@@ -405,52 +401,66 @@ read_ispell (pipe_t * ispell_pipe, char *file, int line)
 	  if (number_lines)
 	    printf ("%d: ", line);
 
-	  for (pos = 2; str->str[pos] != ' '; pos++)
-	    putchar (str->str[pos]);
+	  for (pos = 2; buff[pos] != ' '; pos++)
+	    putchar (buff[pos]);
 	  putchar ('\n');
 
 	  continue;
 	}
 
-      if (str->str[0] == '?' && !verbose)
+      if (b0 == '?' && !verbose)
 	continue;
 
-      error (0, 0, "unrecognized Ispell line `%s'", str_to_nstr (str));
+      error (0, 0, "unrecognized Ispell line `%s'", buff);
     }
+  free(buff);
 }
 
 /* Read from the stderr of the connected process as long as there
    remains data in the channel, and print each error.  Must be called
    from the parent process connected with Ispell by *THE_PIPE (created
    by `new_pipe').  */
+#undef INITIAL_BUFF_SIZE
+#define INITIAL_BUFF_SIZE   256
 
 void
 read_ispell_errors (pipe_t * the_pipe)
 {
   struct timeval time_out;
-  str_t *str = str_make (0);
+  size_t buff_size = INITIAL_BUFF_SIZE;
+  char *buff = xmalloc(sizeof(char[INITIAL_BUFF_SIZE]));
+  ssize_t len;
 
   time_out.tv_sec = time_out.tv_usec = 0;
 
   while (select (FD_SETSIZE, &(the_pipe->error_set), NULL, NULL,
 		 &time_out) == 1)
     {
-      str = str_make (str);
-
-      if (str_add_line_from_desc (str, the_pipe->perr) == ADD_LINE_EOF)
-	/* Ispell closed its stderr.  */
-	error (EXIT_FAILURE, 0, "premature EOF from Ispell's stderr");
+       len = getline(&buff, &buff_size, the_pipe->fperr);
+       if (len == -1) {
+         if (feof(the_pipe->fperr))
+	    /* Ispell closed its stderr.  */
+	    error (EXIT_FAILURE, 0, "premature EOF from Ispell's stderr");
+        else
+            error(EXIT_FAILURE, errno, "error reading errors of ispell/aspell");
+      }
 
       /* Strip the crlf.  */
-      str->len -= 2;
-      str->str[str->len - 1] = 0;
-
-      if (!memcmp (str->str, "Can't open ", strlen ("Can't open ")))
+      if (len > 1  &&  buff[len-1] == '\r'  ||  buff[len-1] == '\n') {
+	if (len > 2  &&  buff[len-2] == '\r'  ||  buff[len-2] == '\n')
+	    len -= 2;
+	else
+            len -= 1;
+	buff[len] = 0;
+      }
+	
+      if (!memcmp (buff, "Can't open ", strlen ("Can't open ")))
 	error (EXIT_FAILURE, 0, "%s: cannot open",
-	       str->str + strlen ("Can't open "));
+	       buff + strlen ("Can't open "));
 
-      fprintf (stderr, "%s: %s\n", ispell_prog, str->str);
+      fprintf (stderr, "%s: %s\n", ispell_prog, buff);
     }
+  free(buff);
 }
 
 /* Create *THE_PIPE, setting up the file descriptors and streams, and
@@ -472,6 +482,9 @@ new_pipe (pipe_t * the_pipe)
     error (EXIT_FAILURE, errno, "error creating pipe to Ispell");
   the_pipe->pin = ifd[0];
   the_pipe->cout = ifd[1];
+  the_pipe->fpin = fdopen(the_pipe->pin, "r");
+  if (the_pipe->fpin == NULL)
+    error (EXIT_FAILURE, errno, "error opening the pipe to Ispell"); 
 
   if (pipe (ofd) < 0)
     error (EXIT_FAILURE, errno, "error creating pipe to Ispell");
@@ -482,6 +495,9 @@ new_pipe (pipe_t * the_pipe)
     error (EXIT_FAILURE, errno, "error creating pipe to Ispell");
   the_pipe->perr = efd[0];
   the_pipe->cerr = efd[1];
+  the_pipe->fperr = fdopen(the_pipe->perr, "r");
+  if (the_pipe->fperr == NULL)
+    error (EXIT_FAILURE, errno, "error opening the error pipe to Ispell");
 
   FD_ZERO (&(the_pipe->error_set));
   FD_SET (the_pipe->perr, &(the_pipe->error_set));
@@ -529,29 +545,34 @@ parent (pipe_t * the_pipe, int argc, char **argv)
      prints it if the flag `--ispell-version' or `-I' was used.
      FIXME: check that the version is high enough that it is going to
      be able to interact with GNU Spell sucessfully.  */
-
   {
-    int pos = 0;
-    str_t *ispell_version = str_make (0);
-    str_t *str = str_make (0);
+    size_t buff_size = INITIAL_BUFF_SIZE;
+    char *buff = xmalloc(sizeof(char[INITIAL_BUFF_SIZE]));
+    ssize_t len;
 
-    if (str_add_line_from_desc (str, the_pipe->pin) == ADD_LINE_EOF)
-      error (EXIT_FAILURE, 0, "premature EOF from Ispell's stdout");
+    size_t ipos = 0;
+    size_t fpos;
 
-    for (; !isdigit (str->str[pos]) && pos <= str->len; pos++);
-    for (; str->str[pos] != ' ' && pos <= str->len; pos++)
-      str_add_char (ispell_version, str->str[pos]);
+    len = getline(&buff, &buff_size, the_pipe->fpin);
+    if (len == -1) {
+      if (feof(the_pipe->fpin))
+          error (EXIT_FAILURE, 0, "premature EOF from Ispell's stdout");
+      else
+          error(EXIT_FAILURE, errno, "error reading errors of ispell/aspell");
+    }
 
-    if (show_ispell_version)
-      {
-	printf ("%s: Ispell version %s\n", program_name,
-		str_to_nstr (ispell_version));
-	exit (EXIT_SUCCESS);
-      }
+    if (show_ispell_version) {
+      for (; !isdigit (buff[ipos]) && ipos < len; ipos++);
+      fpos = ipos;
+      for (; buff[fpos] != ' ' && fpos < len; fpos++);
+      buff[fpos] = 0;
+
+      printf ("%s: Ispell version %s\n", program_name, buff+ipos);
+      exit (EXIT_SUCCESS);
+    }
+    free(buff);
   }
 
-  file = xstrdup ("-");
-
   if (argc == 1)
     read_file (the_pipe, stdin, "-");
 
@@ -563,11 +584,8 @@ parent (pipe_t * the_pipe, int argc, char **argv)
 
       if (file[0] == '-' && file[1] == 0)
 	{
-	  if (!read_stdin)
-	    {
-	      read_stdin = 1;
-	      stream = stdin;
-	    }
+	    read_stdin = 1;
+	    stream = stdin;
 	}
       else
 	{
@@ -626,17 +644,21 @@ run_ispell_in_child (pipe_t * the_pipe)
       error (EXIT_FAILURE, errno, "error duping to stderr");
 
   if (dictionary != NULL)
-    if (execl (ispell_prog, "ispell", "-a", "-p", dictionary, NULL)
-	< 0)
-      error (EXIT_FAILURE, errno, "error executing %s", ispell_prog);
+    if (execlp (ispell_prog, "ispell", "-a", "-p", dictionary, NULL) < 0)
+      if (errno != ENOENT  ||
+	   execlp ("aspell", "aspell", "-a", "-p", dictionary, NULL) < 0)
+	error (EXIT_FAILURE, errno, "error executing ispell/aspell");
 
   if (british)
-    if (execl (ispell_prog, "ispell", "-a", "-d", "british", NULL)
-	< 0)
-      error (EXIT_FAILURE, errno, "error executing %s", ispell_prog);
+    if (execlp (ispell_prog, "ispell", "-a", "-d", "british", NULL) < 0)
+      if (errno != ENOENT  ||
+	   execlp ("aspell", "aspell", "-a", "-d", "british", NULL) < 0)
+        error (EXIT_FAILURE, errno, "error executing ispell/aspell");
 
-  if (execl (ispell_prog, "ispell", "-a", NULL) < 0)
-    error (EXIT_FAILURE, errno, "error executing %s", ispell_prog);
+  if (execlp (ispell_prog, "ispell", "-a", NULL) < 0)
+    if (errno != ENOENT  ||
+         execlp ("aspell", "aspell", "-a", NULL) < 0)
+      error (EXIT_FAILURE, errno, "error executing ispell/aspell");
 }
 
 /* Return a NUL-terminated character string, the meaning of the error
@@ -690,7 +712,7 @@ xmalloc (size_t size)
   void *ptr = malloc (size);
 
   if (!ptr)
-    error (EXIT_FAILURE, 0, "virtual memory exhausted");
+    error (EXIT_FAILURE, errno, "virtual memory exhausted");
   return ptr;
 }
 
@@ -705,21 +727,15 @@ xrealloc (void *ptr, size_t size)
     return xmalloc (size);
   ptr = realloc (ptr, size);
   if (!ptr)
-    error (EXIT_FAILURE, 0, "virtual memory exhausted");
+    error (EXIT_FAILURE, errno, "virtual memory exhausted");
   return ptr;
 }
 
-/* Duplicate STR, returning an identical malloc'd string.  I first
-   just did this for error checking, calling `strdup', but the task is
-   so simple I decided to just do it here--it saves a call.  */
-
 static char *
 xstrdup (const char *str)
 {
-  size_t len = strlen (str) + 1;
-  void *new = xmalloc (len);
-
-  memcpy (new, (void *) str, len);
-
-  return (char *) new;
+  char *ptr = strdup (str);
+  if (!ptr)
+    error (EXIT_FAILURE, errno, "virtual memory exhausted");
+  return ptr;
 }
